#ifdef HAVE_CONFIG_H
#include <libucdaemon_config.h>
#endif

#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <unistd.h>

#include <libucdaemon.h>

/*
 * The daemon status could be sent to the parent either via a signal or via a pipe.
 * With signals there may be a chance for signal to be sent to a wrong process --
 * if the parent exited and another process started with the same PID. So disable
 * signals.
 */
#undef ENABLE_SIGNALS

#define UCD_ARG_EXECED "--ucdaemon-execed"
#define UCD_ARG_PIPE_FD0 "--ucdaemon-pipe-fd0="
#define UCD_ARG_PIPE_FD1 "--ucdaemon-pipe-fd1="

#define UCD_DAEMON_SIGNAL SIGUSR1

static char *ucd_daemon_name;

static unsigned int ucd_daemon_timeout;

static int ucd_detached;

#ifndef ENABLE_SIGNALS
/*
 * Daemon's end of the pipe used to signal status to the parent.
 */
static int ucd_signal_fd = -1;
#endif

#ifndef ENABLE_SIGNALS
static void
ucd_wait_signal(int fd)
{
	int status;

	if (read(fd, &status, sizeof(int)) != sizeof(int)) {
		ucd_error("Could not read status of the daemon process: %s.", strerror(errno));
	}

	if (status == 0) {
		ucd_log(LOG_INFO, "The daemon process returned status %d (success).", status);
		_exit(EXIT_SUCCESS);
	} else {
		ucd_log(LOG_ERR, "The daemon process returned status %d (failure).", status);
		_exit(EXIT_FAILURE);
	}
}
#else
static void
ucd_daemon_signal_handler(int signo, siginfo_t *info, void *context)
{
	int status;

	status = info->si_code == SI_QUEUE ? info->si_value.sival_int : 0;
	if (status == 0) {
		ucd_log(LOG_INFO, "The daemon process returned status %d (success).", status);
		_exit(EXIT_SUCCESS);
	} else {
		ucd_log(LOG_ERR, "The daemon process returned status %d (failure).", status);
		_exit(EXIT_FAILURE);
	}
}
#endif

static void
ucd_alarm_handler(int signo)
{
	ucd_log(LOG_ERR, "Could not receive success notification from the daemon process.");
	_exit(EXIT_FAILURE);
}

static void
ucd_child_handler(int signo)
{
	pid_t pid;
	int status;

	pid = waitpid(-1, &status, WNOHANG);
	if (pid == -1) {
		ucd_error("Could not get status of the terminated daemon process: %s.", strerror(errno));
	}
	if (pid == 0) {
		ucd_error("Could not get status of the terminated daemon process: process not found.");
	}

	if (WIFEXITED(status)) {
		status = WEXITSTATUS(status);
		if (status == 0) {
			ucd_log(LOG_INFO, "The daemon process returned exit code %d (success).", status);
			_exit(EXIT_SUCCESS);
		}
		ucd_log(LOG_ERR, "The daemon process returned exit code %d (failure).", status);

	} else if (WIFSIGNALED(status)) {
		int signo = WTERMSIG(status);
		ucd_log(LOG_ERR, "The daemon process terminated by signal %d.", signo);
	} else {
		ucd_log(LOG_ERR, "The daemon process terminated for unknown reason.");
	}
	_exit(EXIT_FAILURE);
}

static void
ucd_setup_signal_handlers()
{
	struct sigaction sa;

	memset(&sa, 0, sizeof sa);
	sigemptyset(&sa.sa_mask);

#ifdef ENABLE_SIGNALS
	sa.sa_flags = SA_SIGINFO;
	sa.sa_sigaction = ucd_daemon_signal_handler;
	if (sigaction(UCD_DAEMON_SIGNAL, &sa, NULL) < 0) {
		ucd_error("Could not set signal handler.");
	}
#endif

	sa.sa_flags = 0;
	sa.sa_handler = ucd_alarm_handler;
	if (sigaction(SIGALRM, &sa, NULL) < 0) {
		ucd_error("Could not set signal handler.");
	}

	sa.sa_flags = SA_NOCLDSTOP /*| SA_NOCLDWAIT*/;
	sa.sa_handler = ucd_child_handler;
	if (sigaction(SIGCHLD, &sa, NULL) < 0) {
		ucd_error("Could not set signal handler.");
	}
}

static void
ucd_reset_signal_handlers()
{
	/* reset signal handlers */
#ifdef ENABLE_SIGNALS
	signal(UCD_DAEMON_SIGNAL, SIG_DFL);
#endif
	signal(SIGALRM, SIG_DFL);
	signal(SIGCHLD, SIG_DFL);

	/* ignore job control signals */
	signal(SIGTTOU, SIG_IGN);
	signal(SIGTTIN, SIG_IGN);
	signal(SIGTSTP, SIG_IGN);
}

static void
ucd_detach(const char *name, int flags)
{
	if (setsid() < 0) {
		ucd_error("Could not create session: %s", strerror(errno));
	}

	if ((flags & UCD_NOCHDIR) == 0 && chdir("/") < 0) {
		ucd_error("Could not change working directory: %s", strerror(errno));
	}

	openlog(name, LOG_PID, LOG_DAEMON);
	ucd_detached = 1;

	if ((flags & UCD_NOCLOSE) == 0) {
		close(0);
		close(1);
		close(2);

		if (open("/dev/null", O_RDWR) < 0) {
			ucd_error("Could not open /dev/null: %s", strerror(errno));
		}
		if (dup(0) < 0) {
			ucd_error("Could not dup /dev/null: %s", strerror(errno));
		}
		if (dup(0) < 0) {
			ucd_error("Could not dup /dev/null: %s", strerror(errno));
		}
	}

	umask(0);
}

int
ucd_daemon_start(char *name, int flags,
		 int (*parent_init)(int argc, char *argv[]),
		 int (*daemon_main)(int argc, char *argv[]),
		 int argc, char *argv[])
{
	int ret;
	pid_t pid;
#ifndef HAVE_FORK
	int execed;
	int exec_argc;
	char **exec_argv;
#endif
#ifndef ENABLE_SIGNALS
	int pipe_fds[2];
#ifndef HAVE_FORK
	char arg_pipe_fd0[sizeof(UCD_ARG_PIPE_FD0) + 16];
	char arg_pipe_fd1[sizeof(UCD_ARG_PIPE_FD0) + 16];
#endif
#endif

	ucd_detached = 0;

	/* check arguments */
	if (daemon_main == NULL) {
		ucd_error("'main' is null");
	}
	if (argv == NULL || argv[0] == 0) {
		ucd_error("'argv' is null");
	}
	if (argc <= 0) {
		ucd_error("'argc' is less than or equal to zero");
	}

	/* determine the daemon name */
	if (name == NULL) {
		name = strrchr(argv[0], '/');
		if (name == NULL) {
			name = argv[0];
		} else {
			++name;
		}
	}
	ucd_set_name(name);

	/* if daemonizing is suppressed */
	if ((flags & UCD_NODETACH) != 0) {
		if (parent_init != NULL) {
			ret = (*parent_init)(argc, argv);
			if (ret != 0) {
				exit(EXIT_FAILURE);
			}
		}
		return daemon_main(argc, argv);
	}

#ifdef HAVE_FORK
	if (parent_init != NULL) {
		ret = (*parent_init)(argc, argv);
		if (ret != 0) {
			exit(EXIT_FAILURE);
		}
	}

	/* prepare for daemon signal */
	if (ucd_daemon_timeout) {
#ifndef ENABLE_SIGNALS
		if (pipe(pipe_fds) < 0) {
			ucd_error("fork() failed: %s", strerror(errno));
		}
#endif
		ucd_setup_signal_handlers();
	}

	pid = fork();
	if (pid < 0) {
		ucd_error("fork() failed: %s", strerror(errno));
	}

	if (pid == 0) {
		/* child process */
		ucd_detach(ucd_daemon_name, flags);
		ucd_reset_signal_handlers();

#ifndef ENABLE_SIGNALS
		close(pipe_fds[0]);
		ucd_signal_fd = pipe_fds[1];
#endif

		return daemon_main(argc, argv);
	}
#else
	/* check if we re-enter here after vfork/exec */
	execed = (argc > 1 && strcmp(argv[argc - 1], UCD_ARG_EXECED) == 0);
	if (execed) {
		--argc;

#ifndef ENABLE_SIGNALS
		/* check if there are pipe fds */
		pipe_fds[0] = -1;
		pipe_fds[1] = -1;
		if (argc > 1 && strncmp(argv[argc - 1], UCD_ARG_PIPE_FD1, sizeof UCD_ARG_PIPE_FD1 - 1) == 0) {
			--argc;
			pipe_fds[1] = atoi(argv[argc] + sizeof UCD_ARG_PIPE_FD1 - 1);
		}
		if (argc > 1 && strncmp(argv[argc - 1], UCD_ARG_PIPE_FD0, sizeof UCD_ARG_PIPE_FD0 - 1) == 0) {
			--argc;
			pipe_fds[0] = atoi(argv[argc] + sizeof UCD_ARG_PIPE_FD0 - 1);
		}

		close(pipe_fds[0]);
		ucd_signal_fd = pipe_fds[1];
#endif

		ucd_detach(ucd_daemon_name, flags);
		ucd_reset_signal_handlers();

		argv[argc] = NULL;
		return daemon_main(argc, argv);
	}

	if (parent_init != NULL) {
		ret = (*parent_init)(argc, argv);
		if (ret != 0) {
			exit(EXIT_FAILURE);
		}
	}

	/* prepare arguments for exec */
	exec_argc = argc + 1;
	if (ucd_daemon_timeout) {
		exec_argc += 2;
	}
	exec_argv = malloc((exec_argc + 1) * sizeof(char *));
	if (exec_argv == NULL) {
		ucd_error("out of memory");
	}
	memcpy(exec_argv, argv, argc * sizeof(char *));
	if (ucd_daemon_timeout) {
		exec_argv[argc++] = arg_pipe_fd0;
		exec_argv[argc++] = arg_pipe_fd1;
		exec_argv[argc++] = UCD_ARG_EXECED;
	} else {
		exec_argv[argc++] = UCD_ARG_EXECED;
	}
	exec_argv[argc] = NULL;

	/* prepare for daemon signal */
	if (ucd_daemon_timeout) {
#ifndef ENABLE_SIGNALS
		if (pipe(pipe_fds) < 0) {
			ucd_error("fork() failed: %s", strerror(errno));
		}
		snprintf(arg_pipe_fd0, sizeof arg_pipe_fd0, "%s%d", UCD_ARG_PIPE_FD0, pipe_fds[0]);
		snprintf(arg_pipe_fd1, sizeof arg_pipe_fd1, "%s%d", UCD_ARG_PIPE_FD1, pipe_fds[1]);
#endif
		ucd_setup_signal_handlers();
	}

	pid = vfork();
	if (pid < 0) {
		ucd_error("vfork() failed: %s", strerror(errno));
	}
	if (pid == 0) {
		/* child process */
		execv(argv[0], exec_argv);

		/* we can get here only if the exec above fails */
		_exit(EXIT_FAILURE);
	}

	free(exec_argv);
#endif

	/* wait for daemon signal */
	if (ucd_daemon_timeout) {
		alarm(ucd_daemon_timeout);
#ifndef ENABLE_SIGNALS
		close(pipe_fds[1]);
		ucd_wait_signal(pipe_fds[0]);
#else
		pause();
#endif
	}

	exit(EXIT_SUCCESS);
}

const char *
ucd_get_name(void)
{
	return ucd_daemon_name;
}

void
ucd_set_name(char *name)
{
#if 0
	if (ucd_daemon_name != NULL) {
		free(ucd_daemon_name);
	}

	if (name == NULL) {
		ucd_daemon_name = NULL;
	} else {
		ucd_daemon_name = strdup(name);
		if (ucd_daemon_name == NULL) {
			ucd_daemon_name = name;
			ucd_error("out of memory");
		}
	}
#else
	ucd_daemon_name = name;
#endif
}

int
ucd_daemon_is_detached(void)
{
	return ucd_detached;
}

void
ucd_daemon_wait(unsigned int timeout)
{
	ucd_daemon_timeout = timeout;
}

void
ucd_daemon_signal()
{
#ifndef ENABLE_SIGNALS
	ucd_daemon_signal_status(0);
#else
	if (!ucd_detached) {
		ucd_daemon_signal_status(0);
		return;
	}

	kill(getppid(), UCD_DAEMON_SIGNAL);
#endif
}

void
ucd_daemon_signal_status(int status)
{
#ifndef ENABLE_SIGNALS
	struct sigaction sa, osa;
#else
	union sigval sigval;
#endif

	if (!ucd_detached) {
		if (status == 0) {
			ucd_log(LOG_INFO, "The daemon process returned status %d (success).", status);
		} else {
			ucd_log(LOG_ERR, "The daemon process returned status %d (failure).", status);
		}
		return;
	}

#ifndef ENABLE_SIGNALS
	if (ucd_signal_fd < 0) {
		ucd_log(LOG_INFO, "Cannot send status over closed daemon pipe.");
		return;
	}

	/* temporary disable SIGPIPE - the parent may have already exited */
	memset(&sa, 0, sizeof sa);
	sigemptyset(&sa.sa_mask);
	sa.sa_handler = SIG_IGN;
	if (sigaction(SIGPIPE, &sa, &osa) < 0) {
		ucd_error("Could not set signal handler.");
	}

	/* write status to the pipe */
	if (write(ucd_signal_fd, &status, sizeof(int)) != sizeof(int)) {
		ucd_log(LOG_ERR, "The daemon process failed to send its status (%d).", status);
	}

	/* there is no longer need for pipe - close it */
	close(ucd_signal_fd);
	ucd_signal_fd = -1;

	/* restore SIGPIPE */
	if (sigaction(SIGPIPE, &osa, NULL) < 0) {
		ucd_error("Could not set signal handler.");
	}

#else
	sigval.sival_int = status;
	sigqueue(getppid(), UCD_DAEMON_SIGNAL, sigval);
#endif
}

void
ucd_vlog(int priority, const char *format, va_list ap)
{
	if (ucd_detached) {
		vsyslog(priority | LOG_DAEMON, format, ap);
	} else {
		if (ucd_daemon_name) {
			fprintf(stderr, "%s: ", ucd_daemon_name);
		}
		vfprintf(stderr, format, ap);
		fprintf(stderr, "\n");
	}
}

void
ucd_log(int priority, const char *format, ...)
{
	va_list ap;

	va_start(ap, format);
	ucd_vlog(priority, format, ap);
	va_end(ap);
}

void
ucd_error(const char *format, ...)
{
	va_list ap;

	va_start(ap, format);
	ucd_vlog(LOG_ERR, format, ap);
	va_end(ap);

	exit(EXIT_FAILURE);
}
