#ifdef HAVE_CONFIG_H
#include <libucdaemon_config.h>
#endif

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>

#include <libucdaemon.h>

#define UCD_PIDNAMELEN 256

static char *ucd_pidfile_name;

static int ucd_pidfile_fd = -1;

static int
ucd_set_file_lock(int fd, int type)
{
	struct flock lock;

	memset(&lock, 0, sizeof lock);
	lock.l_type = type;

	if (fcntl(fd, F_SETLK, &lock) < 0) {
		return -1;
	}

	return 0;
}

static int
ucd_get_file_lock_pid(int fd, int type, pid_t *pid)
{
	struct flock lock;

	memset(&lock, 0, sizeof lock);
	lock.l_type = type;

	if (fcntl(fd, F_GETLK, &lock) < 0) {
		return -1;
	}

	if (lock.l_type == F_UNLCK) {
		*pid = -1;
	} else {
		*pid = lock.l_pid;
	}

	return 0;
}

pid_t
ucd_pidfile_read(const char *path)
{
	char name[UCD_PIDNAMELEN];
	size_t size;
	int fd;
	char buf[32];
	char *endptr;
	pid_t pid, pid2;

	if (path != NULL) {
		size = snprintf(name, sizeof name, "%s", path);
	} else {
		size = snprintf(name, sizeof name, "/var/run/%s.pid", ucd_get_name());
	}
	if (size >= sizeof name) {
		ucd_error("Too long PID file name");
	}

	fd = open(name, O_RDWR);
	if (fd < 0) {
		return -1;
	}

	size = read(fd, buf, sizeof buf - 1);
	if (size < 0) {
		close(fd);
		ucd_log(LOG_WARNING, "Failed to read PID file: %s", strerror(errno));
		return -1;
	}
	buf[size] = 0;

	errno = 0;
	pid = (pid_t) strtoul(buf, &endptr, 10);
	if (errno != 0 || endptr == buf || endptr[strspn(endptr, " \r\n")] != 0) {
		close(fd);
		ucd_log(LOG_WARNING, "Failed to parse PID file: %d %s");
		return -1;
	}

	if (ucd_get_file_lock_pid(fd, F_WRLCK, &pid2) < 0) {
		close(fd);
		ucd_log(LOG_WARNING, "Failed to get lock owner for the PID file");
		return -1;
	}

	close(fd);

	if (pid2 == -1) {
		ucd_log(LOG_WARNING, "The PID file seems to be stale");
		return -1;
	}
	if (pid2 != pid) {
		ucd_log(LOG_WARNING, "The PID file content mismatches the lock owner");
		return -1;
	}

	return pid;
}

void
ucd_pidfile_lock(const char *path)
{
	char name[UCD_PIDNAMELEN];
	size_t size;
	char buf[32];
	int len;

	if (path != NULL) {
		size = snprintf(name, sizeof name, "%s", path);
	} else {
		size = snprintf(name, sizeof name, "/var/run/%s.pid", ucd_get_name());
	}
	if (size >= sizeof name) {
		ucd_error("Too long PID file name");
	}

	ucd_pidfile_fd = open(name, O_CREAT|O_RDWR, 0644);
	if (ucd_pidfile_fd < 0) {
		ucd_error("Failed to open PID file: %s", strerror(errno));
	}

	if (ucd_set_file_lock(ucd_pidfile_fd, F_WRLCK) < 0) {
		ucd_error("Failed to lock PID file: %s", strerror(errno));
	}

	ucd_pidfile_name = strdup(name);
	if (ucd_pidfile_name == NULL) {
		unlink(name);
		ucd_error("out of memory");
	}

	snprintf(buf, sizeof buf, "%lu\n", (unsigned long) getpid());
	len = strlen(buf);

	if (write(ucd_pidfile_fd, buf, len) != len) {
		ucd_pidfile_unlock();
		ucd_error("Failed to write PID file: %s", strerror(errno));
	}
	if (ftruncate(ucd_pidfile_fd, len) < 0) {
		ucd_pidfile_unlock();
		ucd_error("Failed to truncate PID file: %s", strerror(errno));
	}
}

void
ucd_pidfile_unlock()
{
	if (ucd_pidfile_name != NULL) {
		if (unlink(ucd_pidfile_name) < 0) {
			ucd_log(LOG_ERR, "Failed to remove PID file: %s", strerror(errno));
		}

		free(ucd_pidfile_name);
		ucd_pidfile_name = NULL;
	}

	if (ucd_pidfile_fd >= 0) {
		ucd_set_file_lock(ucd_pidfile_fd, F_UNLCK);
		close(ucd_pidfile_fd);
		ucd_pidfile_fd = -1;
	}
}
