#ifdef HAVE_CONFIG_H
#include <libucdaemon_config.h>
#endif

#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <libucdaemon.h>

/* TODO: autoconf this value */
#define UCD_NSIG 64

typedef struct ucd_sigpipe_signal_t
{
	int signo;
	int count;
} ucd_sigpipe_signal_t;

struct ucd_sigpipe_t
{
	sigset_t set;
	ucd_sigpipe_t *next;
	int pipe[2];
	int nsig, in, out;
	ucd_sigpipe_signal_t queue[1];
};

/* Global list of pipes. Beware that while modifying the list all signals
   has to be blocked. */
static ucd_sigpipe_t *ucd_sigpipe_list;

static void
ucd_sigpipe_signal_handler(int signo)
{
	ucd_sigpipe_t *sigpipe;
	int in;

	/* look for the pipe associated with signal */
	sigpipe = ucd_sigpipe_list;
	for (;;) {
		if (sigpipe == NULL) {
			/* this should never happen */
			return;
		}
		if (sigismember(&sigpipe->set, signo)) {
			break;
		}
		sigpipe = sigpipe->next;
	}

	/* check if this is a repeated signal in which case just increment
	   the signal count */
	in = sigpipe->out;
	while (in != sigpipe->in) {
		if (sigpipe->queue[in].signo == signo) {
			++sigpipe->queue[in].count;
			return;
		}
		if (++in == sigpipe->nsig) {
			in = 0;
		}
	}

	/* get the next available queue slot */
	in = sigpipe->in + 1;
	if (in == sigpipe->nsig) {
		in = 0;
	}

	if (in == sigpipe->out) {
		/* this should never happen as repeated signals are merged
		   into one slot and we pre-allocate a slot for each enabled
		   signal */
		ucd_error("signal queue overflow");
	}

	/* put the signal into queue */
	sigpipe->queue[in].signo = signo;
	sigpipe->queue[in].count = 1;
	sigpipe->in = in;

	write(sigpipe->pipe[1], "x", 1);
}

static void
ucd_sigpipe_fcntl(int fd)
{
	  int flags;

	  /* set close-on-exec flag */
	  flags = fcntl(fd, F_GETFD, 0);
	  if (flags < 0) {
		  ucd_error("Failed fcntl call: %s", strerror(errno));
	  }
	  flags |= FD_CLOEXEC;
	  if (fcntl(fd, F_SETFD, flags) < 0) {
		  ucd_error("Failed fcntl call: %s", strerror(errno));
	  }

	  /* set non-blocking mode */
	  flags = fcntl(fd, F_GETFL, 0);
	  if (flags < 0) {
		  ucd_error("Failed fcntl call: %s", strerror(errno));
	  }
	  flags |= O_NONBLOCK;
	  if (fcntl (fd, F_SETFL, flags) < 0) {
		  ucd_error("Failed fcntl call: %s", strerror(errno));
	  }
}

ucd_sigpipe_t *
ucd_sigpipe_open(sigset_t *set)
{
	ucd_sigpipe_t *sigpipe;
	int nsig, signo, size, ismember;
	sigset_t allset, oldset;
	struct sigaction sa;

	if (set == NULL) {
		return NULL;
	}

	/* count the number of signals associated with the pipe */
	nsig = 0;
	for (signo = 1; signo <= UCD_NSIG; signo++) {
		ismember = sigismember(set, signo);
		if (ismember < 0) {
			/* presumably the number of signals supported
			   by the system is less than UCD_NSIG */
			break;
		}

		if (ismember > 0) {
			/* check if the signal is already associated with
			   another pipe */
			sigpipe = ucd_sigpipe_list;
			while (sigpipe != NULL) {
				if (sigismember(&sigpipe->set, signo)) {
					ucd_error("A signal can only be associated with single pipe");
				}
				sigpipe = sigpipe->next;
			}

			++nsig;
		}
	}
	if (nsig == 0) {
		return NULL;
	}

	/* allocate and initialize ucd_sigpipe_t struct along with signal queue */
	size = sizeof(ucd_sigpipe_t) + (nsig - 1) * sizeof(ucd_sigpipe_signal_t);
	sigpipe = malloc(size);
	if (sigpipe == NULL) {
		ucd_error("out of memory");
	}
	memset(sigpipe, 0, size);
	memcpy(&sigpipe->set, set, sizeof(sigset_t));
	sigpipe->nsig = nsig;

	/* open the pipe */
	if (pipe(sigpipe->pipe) < 0) {
		ucd_error("Failed to create a signal pipe");
	}
	ucd_sigpipe_fcntl(sigpipe->pipe[0]);
	ucd_sigpipe_fcntl(sigpipe->pipe[1]);

	/* temporary block all signals */
	sigfillset(&allset);
	if (sigprocmask(SIG_SETMASK, &allset, &oldset) < 0) {
		ucd_error("Failed to set signal mask");
	}

	/* update the global list */
	sigpipe->next = ucd_sigpipe_list;
	ucd_sigpipe_list = sigpipe;

	/* unblock the signals */
	sigprocmask(SIG_SETMASK, &oldset, NULL);

	/* finally setup the handler for each signal */
	memset(&sa, 0, sizeof sa);
	sa.sa_handler = ucd_sigpipe_signal_handler;
	memcpy(&sa.sa_mask, set, sizeof(sigset_t));
	sa.sa_flags = 0;
	for (signo = 1; signo <= UCD_NSIG; signo++) {
		if (sigismember(set, signo) > 0) {
			if (sigaction(signo, &sa, NULL) < 0) {
				ucd_error("Could not set signal handler.");
			}
		}
	}

	return sigpipe;
}

int
ucd_sigpipe_close(ucd_sigpipe_t *sigpipe)
{
	ucd_sigpipe_t **psigpipe;
	int signo;
	sigset_t allset, oldset;

	if (sigpipe == NULL) {
		return -1;
	}

	/* remove the handler for each signal */
	for (signo = 1; signo <= UCD_NSIG; signo++) {
		if (sigismember(&sigpipe->set, signo) > 0) {
			signal(signo, SIG_DFL);
		}
	}

	/* temporary block all signals */
	sigfillset(&allset);
	if (sigprocmask(SIG_SETMASK, &allset, &oldset) < 0) {
		ucd_error("Failed to set signal mask");
	}

	/* update the global list */
	psigpipe = &ucd_sigpipe_list;
	while (*psigpipe != NULL) {
		if (*psigpipe == sigpipe) {
			*psigpipe = sigpipe->next;
			break;
		}
		psigpipe = &(*psigpipe)->next;
	}

	/* unblock the signals */
	sigprocmask(SIG_SETMASK, &oldset, NULL);

	/* release resources */
	close(sigpipe->pipe[0]);
	close(sigpipe->pipe[1]);
	free(sigpipe);

	return 0;
}

int
ucd_sigpipe_get_fd(ucd_sigpipe_t *sigpipe)
{
	if (sigpipe == NULL) {
		return -1;
	}

	return sigpipe->pipe[0];
}

int
ucd_sigpipe_get_signal(ucd_sigpipe_t *sigpipe, int *pcount)
{
	sigset_t oldset;
	int signo, count;
	char dummy;

	if (sigpipe == NULL) {
		return -1;
	}

	/* clean the pipe, the number of bytes in the pipe
	   may differ from the queue size but after some
	   number of calls they both should become empty */
	read(sigpipe->pipe[0], &dummy, 1);

	/* temporary block the signals that go to the queue */
	if (sigprocmask(SIG_BLOCK, &sigpipe->set, &oldset)) {
		return -1;
	}

	/* get the next signal from the queue if any */
	if (sigpipe->out == sigpipe->in) {
		signo = 0;
		count = 0;
	} else {
		signo = sigpipe->queue[sigpipe->out].signo;
		count = sigpipe->queue[sigpipe->out].count;

		if (++sigpipe->out == sigpipe->nsig) {
			sigpipe->out = 0;
		}
	}

	/* unblock the signals */
	sigprocmask(SIG_SETMASK, &oldset, NULL);

	/* return the data just taken from the queue */
	if (pcount != NULL) {
		*pcount = count;
	}
	return signo;
}
